# File Emsao/emming.x
# February 27, 1997
# Modified by Doug Mink, Harvard-Smithsonian Center for Astrophysics
# Most of this code is from noao.onedspec.splot's spdeblend.x, voigt.x,
# and deblend.x

include	<error.h>
include	<gset.h>
include "emv.h"

define	NSUB	3	# Number of pixel subsamples
define	MC_N	50	# Monte-Carlo samples
define	MC_P	10	# Percent done interval (percent)
define	MC_SIG	68	# Sigma sample point (percent)

# EMMING -- Deblend lines in a spectral region. 

procedure emming (pix, n, pixleft, pixright,
		  ltype, ng, gcenter, gheight, gwidth,
		  nparams, debug, fitvec, covar, ierr)

real	pix[ARB]		# Spectrum data
int	n			# Number of points
int	pixleft			# First pixel to fit
int	pixright		# Last pixel to fit
int	ltype			# Profile (gaussian, lorentzian, voigt)
int	ng			# Number of components
double	gcenter[ARB]		# Centers of line fits in wavelength
double	gheight[ARB]		# Heights of line fits
double	gwidth[ARB]		# Widths of line fits
int	nparams			# Number of parameters to fit
bool	debug			# True for extra logging
double	fitvec[ARB]		# Results of fit: continuum, lines
double	covar[ARB]		# Errors in fit: continuum, lines
int	ierr			# Error code if fit not completed

real	wx1, wy1		# Wavelength and pixel value at first pixel
real	wx2, wy2		# Wavelength and pixel value at  last pixel
pointer	xg, yg, sg, lg		# Pointers to fit parameters
int	fit[5], nsub, mc_p, mc_sig, mc_n
int	i, j, i1, npts, nlines, maxlines, key, ifit
long	seed
double	w, dw, wyc, u, v
double	slope, peak, flux, cont, gfwhm, lfwhm, eqw, chisq, rms
double	sigma0, invgain, wyc1, slope1, flux1, cont1, eqw1
real	scale, sscale
double	dscale
bool	fitit, fitg, fitl
pointer	xg1, yg1, sg1, lg1
pointer	sp, cmd, x, y, s, z, waves, gfwhms, lfwhms, peaks, ym
pointer	conte, xge, yge, sge, lge, fluxe, eqwe

int	clgeti()
real	model(), gasdev(), asumr()
double	clgetd(), wcs_p2w(), wcs_w2p()
errchk	dofit, dorefit

define	done_	99

begin
	call smark (sp)
	call salloc (cmd, SZ_FNAME, TY_CHAR)

	# Set up start and end points for continuum
	j = pixright
	i1 = pixleft
	npts = j - i1 + 1
	if (npts < 3) {
	    call eprintf ("At least 3 points are required\n")
	    call sfree (sp)
	    return
	    }

	wx1 = wcs_p2w (double (pixleft))
	wx2 = wcs_w2p (double (pixright))
	wy1 = pix[pixleft]
	wy2 = pix[pixright]

	# Allocate space for the points to be fit.
	call salloc (x, npts, TY_REAL)
	call salloc (y, npts, TY_REAL)
	call salloc (s, npts, TY_REAL)
	call salloc (z, npts, TY_REAL)

	# Scale the data.
	sigma0 = clgetd ("sigma0")
	invgain = clgetd ("invgain")
	mc_n = clgeti ("nerrsample")
	if (IS_INDEF(sigma0) || IS_INDEF(invgain) || sigma0<0. ||
	    invgain<0. || (sigma0 == 0. && invgain == 0.)) {
	    sigma0 = INDEF
	    invgain = INDEF
	    }
	scale = 0.
	do i = 1, npts {
	    Memr[x+i-1] = wcs_p2w (double (i1+i-1))
	    Memr[y+i-1] = pix[i1+i-1]
	    if (Memr[y+i-1] <= 0.) {
		if (invgain != 0.) {
		    sigma0 = INDEF
		    invgain = INDEF
		    call eprintf (
			"WARNING: Cannot compute errors with non-zero gain")
		    call eprintf (
			" and negative pixel values.\n")
		    }
		}
	    scale = max (scale, abs (Memr[y+i-1]))
	    }
	if (IS_INDEF(sigma0)) {
	    call amovkr (1., Memr[s], npts)
	    sscale = 1.
	    }
	else {
	    do i = 1, npts
		Memr[s+i-1] = sqrt (sigma0 ** 2 + invgain * Memr[y+i-1])
	    sscale = asumr (Memr[s], npts) / npts
	    }
	dscale = double (scale)
	call adivkr (Memr[y], scale, Memr[y], npts)
	call adivkr (Memr[s], sscale, Memr[s], npts)

	# Select the lines to be fit.  If no lines return.
	fitit = false
	fitg = false
	fitl = false
	maxlines = ng
	call malloc (waves, ng, TY_DOUBLE)
	call malloc (peaks, ng, TY_DOUBLE)
	call malloc (gfwhms, ng, TY_DOUBLE)
	call malloc (lfwhms, ng, TY_DOUBLE)
	nlines = 0

	#Lines (Gaussian, Lorentzian, Voigt)
	do i = 1, ng {
	    switch (ltype) {
		case GAUSS:
		    peak = INDEF
		    gfwhm = INDEF
		    lfwhm = INDEF
		    fitg = TRUE
		case LORENTZ:
		    peak = INDEF
		    gfwhm = INDEF
		    lfwhm = INDEF
		    fitl = TRUE
		case VOIGT:
		    peak = INDEF
		    gfwhm = INDEF
		    lfwhm = INDEF
		    fitg = TRUE
		    fitl = TRUE
		}
	    Memd[waves+i] = wcs_p2w (gcenter[i])
	    Memd[peaks+i] = gheight[i]
	    Memd[gfwhms+i] = gwidth[i]
	    Memd[lfwhms+i] = gwidth[i]
	    nlines = nlines + 1
	    }
	if (nlines == 0)
	    goto done_

	# Allocate memory.
	if (ng == 0) {
	    call malloc (xg, nlines, TY_DOUBLE)
	    call malloc (yg, nlines, TY_DOUBLE)
	    call malloc (sg, nlines, TY_DOUBLE)
	    call malloc (lg, nlines, TY_DOUBLE)
	    }
	else if (ng != nlines) {
	    call realloc (xg, nlines, TY_DOUBLE)
	    call realloc (yg, nlines, TY_DOUBLE)
	    call realloc (sg, nlines, TY_DOUBLE)
	    call realloc (lg, nlines, TY_DOUBLE)
	    }
	ng = nlines

	# Do fits.
	fit[3] = 3

	do ifit = 1, ng {

	    # Fit Gaussian widths
	    if (fitg)
		fit[4] = 3

	    # Fit Lorentzian widths
	    if (fitl)
		fit[5] = 3

	    # Fit background
	    fit[1] = 2
	    if (debug) {
		call printf ("Fitting...")
		call flush (STDOUT)
		}

	    # Setup initial estimates.
	    if (ifit == 0) {
		slope = (wy2-wy1) / (wx2-wx1) / dscale
		wyc = wy1 / dscale - slope * wx1
		eqw = abs (Memr[x+npts-1] - Memr[x]) / nlines
		do i = 0, nlines-1 {
		    w = Memd[waves+i]
		    peak = Memd[peaks+i]
		    gfwhm = Memd[gfwhms+i]
		    lfwhm = Memd[lfwhms+i]
		    j = max (1, min (n, nint (wcs_w2p (double(w)))))
		    Memd[xg+i] = w
		    if (IS_INDEF(peak))
			Memd[yg+i] = pix[j] / dscale - wyc - slope * w
		    else
			Memd[yg+i] = peak / dscale
		    Memd[sg+i] = 0.
		    Memd[lg+i] = 0.
		    switch (ltype) {
			case GAUSS:
			    if (IS_INDEF(gfwhm))
				Memd[sg+i] = 0.3 * eqw
			    else
				Memd[sg+i] = gfwhm
			case LORENTZ:
			    if (IS_INDEF(lfwhm))
				Memd[lg+i] = 0.3 * eqw
			    else
				Memd[lg+i] = lfwhm
			case VOIGT:
			    if (IS_INDEF(Memd[gfwhms+i]))
				Memd[sg+i] = 0.1 * eqw
			    else
				Memd[sg+i] = gfwhm
			    if (IS_INDEF(Memd[lfwhms+i]))
				Memd[lg+i] = 0.1 * eqw
			    else
				Memd[lg+i] = lfwhm
			}
		    }
		}
	    else {
		call adivkd (Memd[yg], dscale, Memd[yg], ng)
		slope = slope / dscale
		wyc = wyc / dscale - slope * wx1
		}

	    nsub = NSUB
	    dw = (wcs_p2w (double (n)) - wcs_p2w (double (1))) / double (n - 1)
	    iferr (call dofit (fit, Memr[x], Memr[y], Memr[s], npts,
		dw, nsub, wyc, slope,
		Memd[xg], Memd[yg], Memd[sg], Memd[lg], ltype, ng, chisq)) {
		call erract (EA_WARN)
		next
	        }

	    # Compute Monte-Carlo errors.
	    if (mc_n > 9 && !IS_INDEF(sigma0)) {
		mc_p = nint (mc_n * MC_P / 100.)
		mc_sig = nint (mc_n * MC_SIG / 100.)

		call salloc (ym, npts, TY_DOUBLE)
		call salloc (xg1, ng, TY_DOUBLE)
		call salloc (yg1, ng, TY_DOUBLE)
		call salloc (sg1, ng, TY_DOUBLE)
		call salloc (lg1, ng, TY_DOUBLE)
		call salloc (conte, mc_n*ng, TY_DOUBLE)
		call salloc (xge, mc_n*ng, TY_DOUBLE)
		call salloc (yge, mc_n*ng, TY_DOUBLE)
		call salloc (sge, mc_n*ng, TY_DOUBLE)
		call salloc (lge, mc_n*ng, TY_DOUBLE)
		call salloc (fluxe, mc_n*ng, TY_DOUBLE)
		call salloc (eqwe, mc_n*ng, TY_DOUBLE)
		do i = 1, npts {
		    w = Memr[x+i-1]
		    Memd[ym+i-1] = model (w, dw, nsub, Memd[xg], Memd[yg],
			Memd[sg], Memd[lg], ltype, ng) + wyc + slope * w
		    }
		seed = 1
		do i = 0, mc_n-1 {
		    if (i > 0 && mod (i, mc_p) == 0) {
			call printf ("%2d ")
			    call pargi (100 * i / mc_n)
			call flush (STDOUT)
			}
		    do j = 1, npts
			Memr[y+j-1] = Memd[ym+j-1] +
			    sscale / scale * Memr[s+j-1] * gasdev (seed)
		    wyc1 = wyc
		    slope1 = slope
		    call amovd (Memd[xg], Memd[xg1], ng)
		    call amovd (Memd[yg], Memd[yg1], ng)
		    call amovd (Memd[sg], Memd[sg1], ng)
		    call amovd (Memd[lg], Memd[lg1], ng)
		    call dorefit (fit, Memr[x], Memr[y], Memr[s], npts,
			dw, nsub, wyc1, slope1,
			Memd[xg1], Memd[yg1], Memd[sg1], Memd[lg1], ltype,
			ng, chisq)

		    do j = 0, ng-1 {
			cont = wyc + slope * Memd[xg+j]
			cont1 = wyc1 + slope1 * Memd[xg+j]
			switch (ltype) {
			    case GAUSS:
				flux = 1.064467 * Memd[yg+j] * Memd[sg+j]
				flux1 = 1.064467 * Memd[yg1+j] * Memd[sg1+j]
			    case LORENTZ:
				flux = 1.570795 * Memd[yg+j] * Memd[lg+j]
				flux1 = 1.570795 * Memd[yg1+j] * Memd[lg1+j]
			    case VOIGT:
				call voigt (0., 0.832555*Memd[lg+j]/Memd[sg+j],
					    v, u)
				flux = 1.064467 * Memd[yg+j] * Memd[sg+j] / v
				call voigt (0., 0.832555*Memd[lg1+j]/Memd[sg1+j],
					    v, u)
				flux1 = 1.064467 * Memd[yg1+j] * Memd[sg1+j] / v
			    }
			if (cont > 0. && cont1 > 0.) {
			    eqw = -flux / cont
			    eqw1 = -flux1 / cont1
			    }
			else {
			    eqw = 0.
			    eqw1 = 0.
			    }
			Memd[conte+j*mc_n+i] = abs (cont1 - cont)
			Memd[xge+j*mc_n+i] = abs (Memd[xg1+j] - Memd[xg+j])
			Memd[yge+j*mc_n+i] = abs (Memd[yg1+j] - Memd[yg+j])
			Memd[sge+j*mc_n+i] = abs (Memd[sg1+j] - Memd[sg+j])
			Memd[lge+j*mc_n+i] = abs (Memd[lg1+j] - Memd[lg+j])
			Memd[fluxe+j*mc_n+i] = abs (flux1 - flux)
			Memd[eqwe+j*mc_n+i] = abs (eqw1 - eqw)
			}
		    }
		do j = 0, ng-1 {
		    call asrtd (Memd[conte+j*mc_n], Memd[conte+j*mc_n], mc_n)
		    call asrtd (Memd[xge+j*mc_n], Memd[xge+j*mc_n], mc_n)
		    call asrtd (Memd[yge+j*mc_n], Memd[yge+j*mc_n], mc_n)
		    call asrtd (Memd[sge+j*mc_n], Memd[sge+j*mc_n], mc_n)
		    call asrtd (Memd[lge+j*mc_n], Memd[lge+j*mc_n], mc_n)
		    call asrtd (Memd[fluxe+j*mc_n], Memd[fluxe+j*mc_n], mc_n)
		    call asrtd (Memd[eqwe+j*mc_n], Memd[eqwe+j*mc_n], mc_n)
		    }
		call amulkd (Memd[conte], dscale, Memd[conte], mc_n*ng)
		call amulkd (Memd[yge], dscale, Memd[yge], mc_n*ng)
		call amulkd (Memd[fluxe], dscale, Memd[fluxe], mc_n*ng)
		}

	    call amulkd (Memd[yg], dscale, Memd[yg], ng)
	    wyc = (wyc + slope * wx1) * dscale
	    slope = slope * dscale

	    # Compute model spectrum with continuum
	    fitit = true
	    rms = 0.
	    do i = 1, npts {
		w = Memr[x+i-1]
	        Memr[z+i-1] = model (w, dw, nsub, Memd[xg], Memd[yg],
		    Memd[sg], Memd[lg], ltype, ng)
	        Memr[z+i-1] = Memr[z+i-1] + wyc + slope * (w - wx1)
		rms = rms + (Memr[z+i-1] / scale - Memr[y+i-1]) ** 2
		}

	    # Print computed values
	    i = 1
	    key = ''
	    do i = 1, nlines {
		if (debug) {
		    call printf ("\nrms = %8.4g")
			call pargd (dscale * sqrt (chisq / double (npts)))
		    }
		w = Memd[xg+i-1]
		cont = wyc + slope * (w - wx1)
		peak = Memd[yg+i-1]
		gfwhm = Memd[sg+i-1]
		lfwhm = Memd[lg+i-1]
		switch (ltype) {
		    case GAUSS:
			flux = 1.064467 * peak * gfwhm
			if (cont > 0.)
			    eqw = -flux / cont
			else
			    eqw = INDEF
			if (debug) {
			    call printf (
	    "\n%d: center = %8.6g, flux = %8.4g, eqw = %6.4g, gfwhm = %6.4g")
				call pargi (i)
				call pargd (w)
				call pargd (flux)
				call pargd (eqw)
				call pargd (gfwhm)
			    }
		    case LORENTZ:
			flux = 1.570795 * peak * lfwhm
			if (cont > 0.)
			    eqw = -flux / cont
			else
			    eqw = INDEF
			if (debug) {
			    call printf (
	    "\n%d: center = %8.6g, flux = %8.4g, eqw = %6.4g, lfwhm = %6.4g")
				call pargi (i)
				call pargd (w)
				call pargd (flux)
				call pargd (eqw)
				call pargd (lfwhm)
			    }
		    case VOIGT:
			call voigt (0., 0.832555*lfwhm/gfwhm, v, u)
			flux = 1.064467d0 * peak * gfwhm / v
			if (cont > 0.)
			    eqw = -flux / cont
			else
			    eqw = INDEF
			if (debug) {
			    call printf (
	    "\n%d: center = %8.6g, eqw = %6.4g, gfwhm = %6.4g, lfwhm = %6.4g")
				call pargi (i)
				call pargd (w)
				call pargd (eqw)
				call pargd (gfwhm)
				call pargd (lfwhm)
			    }
		    }
		}
	    }

done_
	if (debug)
	    call printf ("Deblending done\n")

	# Log computed values
	if (fitit) {
	    do i = 1, nlines {
		w = Memd[xg+i-1]
		cont = wyc + slope * (w - wx1)
		peak = Memd[yg+i-1]
		gfwhm = Memd[sg+i-1]
		lfwhm = Memd[lg+i-1]
		switch (ltype) {
		    case GAUSS:
			flux = 1.064467 * peak * gfwhm
		    case LORENTZ:
			flux = 1.570795 * peak * lfwhm
		    case VOIGT:
			call voigt (0., 0.832555*lfwhm/gfwhm, v, u)
			flux = 1.064467 * peak * gfwhm / v
		    }

		if (cont > 0.)
		    eqw = -flux / cont
		else
		    eqw = INDEF
		if (debug) {
		    call printf (" %9.7g %9.7g %9.6g %9.4g %9.6g %9.4g %9.4g\n")
			call pargd (w)
			call pargd (cont)
			call pargd (flux)
			call pargd (eqw)
			call pargd (peak)
			call pargd (gfwhm)
			call pargd (lfwhm)
		    }
		if (mc_n > 9 && !IS_INDEF(sigma0)) {
		    if (debug) {
			call printf ("  (%7.5g) (%7.5g) (%7.4g) (%7.4g) (%7.4g) (%7.4g) (%7.4g)\n")
			    call pargd (Memd[xge+(i-1)*mc_n+mc_sig])
			    call pargd (Memd[conte+(i-1)*mc_n+mc_sig])
			    call pargd (Memd[fluxe+(i-1)*mc_n+mc_sig])
			    call pargd (Memd[eqwe+(i-1)*mc_n+mc_sig])
			    call pargd (Memd[yge+(i-1)*mc_n+mc_sig])
			    call pargd (Memd[sge+(i-1)*mc_n+mc_sig])
			    call pargd (Memd[lge+(i-1)*mc_n+mc_sig])
			}
		    }
		}
	    }
	else {
	    call mfree (xg, TY_DOUBLE)
	    call mfree (yg, TY_DOUBLE)
	    call mfree (sg, TY_DOUBLE)
	    call mfree (lg, TY_DOUBLE)
	    ng = 0
	    }

	call mfree (waves, TY_DOUBLE)
	call mfree (peaks, TY_DOUBLE)
	call mfree (gfwhms, TY_DOUBLE)
	call mfree (lfwhms, TY_DOUBLE)
	call sfree (sp)
end



# GASDEV -- Return a normally distributed deviate with zero mean and unit
# variance.  The method computes two deviates simultaneously.
#
# Based on Numerical Recipes by Press, Flannery, Teukolsky, and Vetterling.
# Used by permission of the authors.
# Copyright(c) 1986 Numerical Recipes Software.

real procedure gasdev (seed)

long	seed		# Seed for random numbers

real	v1, v2, r, fac, urand()
int	iset
data	iset/0/

begin
	if (iset == 0) {
	    repeat {
	        v1 = 2 * urand (seed) - 1.
	        v2 = 2 * urand (seed) - 1.
	        r = v1 ** 2 + v2 ** 2
	    } until ((r > 0) && (r < 1))
	    fac = sqrt (-2. * log (r) / r)

	    iset = 1
	    return (v1 * fac)
	} else {
	    iset = 0
	    return (v2 * fac)
	}
end
# VOIGT -- Compute the real (Voigt function) and imaginary parts of the
# complex function w(z)=exp(-z**2)*erfc(-i*z) in the upper half-plane
# z=x+iy.  The maximum relative error of the real part is 2E-6 and the
# imaginary part is 5E-6.
#
# From: Humlicek, J. Quant. Spectrosc. Radiat. Transfer, V21, p309, 1979.

procedure voigt (xarg, yarg, wr, wi)

real	xarg		#I Real part of argument
real	yarg		#I Imaginary part of argument
real	wr		#O Real part of function
real	wi		#O Imaginary part of function

int	i
real	x, y, y1, y2, y3, d, d1, d2, d3, d4, r, r2
real	t[6], c[6], s[6]

data	t/.314240376,.947788391,1.59768264,2.27950708,3.02063703,3.8897249/
data	c/1.01172805,-.75197147,1.2557727e-2,1.00220082e-2,-2.42068135e-4,
	  5.00848061e-7/
data	s/1.393237,.231152406,-.155351466,6.21836624e-3,9.19082986e-5,
	  -6.27525958e-7/

begin
	x = xarg
	y = abs (yarg)
	wr = 0.
	wi = 0.
	y1 = y + 1.5
	y2 = y1 * y1

	# Region II
	if (y < 0.85 && abs(x) > 18.1*y+1.65) {
	    if (abs(x) < 12)
		wr = exp (-x * x)
	    y3 = y + 3
	    do i = 1, 6 {
		r = x - t[i]
		r2 = r * r
		d = 1 / (r2 + y2)
		d1 = y1 * d
		d2 = r * d
		wr = wr + y * (c[i] * (r * d2 - 1.5 * d1) + s[i] * y3 * d2) /
		    (r2 + 2.25)
		r = x + t[i]
		r2 = r * r
		d = 1 / (r2 + y2)
		d3 = y1 * d
		d4 = r * d
		wr = wr + y * (c[i] * (r * d4 - 1.5 * d3) - s[i] * y3 * d4) /
		    (r2 + 2.25)
		wi = wi + c[i] * (d2 + d4) + s[i] * (d1 - d3)
	    }

	# Region I
	} else {
	    do i = 1, 6 {
		r = x - t[i]
		d = 1 / (r * r + y2)
		d1 = y1 * d
		d2 = r * d
		r = x + t[i]
		d = 1 / (r * r + y2)
		d3 = y1 * d
		d4 = r * d
		wr = wr + c[i] * (d1 + d3) - s[i] * (d2 - d4)
		wi = wi + c[i] * (d2 + d4) + s[i] * (d1 - d3)
	    }
	}
end
include	<math.h>
include	<mach.h>

# Elements of fit array.
define	BKG		1	# Background
define	POS		2	# Position
define	INT		3	# Intensity
define	GAU		4	# Gaussian FWHM
define	LOR		5	# Lorentzian FWHM

# Type of constraints.
define	FIXED		1	# Fixed parameter
define	SINGLE		2	# Fit a single value for all lines
define	INDEP		3	# Fit independent values for all lines


# DOFIT -- Fit line profiles.  This is an interface to DOFIT1
# which puts parameters into the required form and vice-versa.
# It also implements a constrained approach to the solution.

procedure dofit (fit, x, y, s, npts, dx, nsub, y1, dy,
    xp, yp, gp, lp, tp, np, chisq)

int	fit[5]		# Fit constraints
real	x[npts]		# X data
real	y[npts]		# Y data
real	s[npts]		# Sigma data
int	npts		# Number of points
real	dx		# Pixel size
int	nsub		# Number of subpixels
real	y1		# Continuum offset
real	dy		# Continuum slope
real	xp[np]		# Profile positions
real	yp[np]		# Profile intensities
real	gp[np]		# Profile Gaussian FWHM
real	lp[np]		# Profile Lorentzian FWHM
int	tp[np]		# Profile type	
int	np		# Number of profiles
real	chisq		# Chi squared

int	i, j, fit1[5]
pointer	sp, a, b
errchk	dofit1

begin
	call smark (sp)
	call salloc (a, 8 + 5 * np, TY_DOUBLE)

	# Convert positions and widths relative to first component.
	Memd[a] = dx
	Memd[a+1] = nsub
	Memd[a+2] = y1
	Memd[a+3] = dy
	Memd[a+4] = yp[1]
	Memd[a+5] = xp[1]
	Memd[a+6] = gp[1]
	Memd[a+7] = lp[1]
	do i = 1, np {
	    b = a + 5 * i + 3
	    Memd[b]   = yp[i] / Memd[a+4]
	    Memd[b+1] = xp[i] - Memd[a+5]
	    switch (tp[i]) {
	    case GAUSS:
		if (Memd[a+6] == 0.)
		    Memd[a+6] = gp[i]
		Memd[b+2] = gp[i] / Memd[a+6]
	    case LORENTZ:
		if (Memd[a+7] == 0.)
		    Memd[a+7] = lp[i]
		Memd[b+3] = lp[i] / Memd[a+7]
	    case VOIGT:
		if (Memd[a+6] == 0.)
		    Memd[a+6] = gp[i]
		Memd[b+2] = gp[i] / Memd[a+6]
		if (Memd[a+7] == 0.)
		    Memd[a+7] = lp[i]
		Memd[b+3] = lp[i] / Memd[a+7]
	    }
	    Memd[b+4] = tp[i]
	}

	# Do fit.
	fit1[INT] = fit[INT]
	do i = 1, fit[BKG] {
	    fit1[BKG] = i
	    fit1[GAU] = min (SINGLE, fit[GAU])
	    fit1[LOR] = min (SINGLE, fit[LOR])
	    do j = FIXED, fit[POS] {
		fit1[POS] = j
		if (np > 1 || j != INDEP)
		    call dofit1 (fit1, x, y, s, npts, Memd[a], np, chisq)
	    }
	    if (np > 1 && (fit[GAU] == INDEP || fit[LOR] == INDEP)) {
		fit1[GAU] = fit[GAU]
		fit1[LOR] = fit[LOR]
		call dofit1 (fit1, x, y, s, npts, Memd[a], np, chisq)
	    }
	}

	y1 = Memd[a+2]
	dy = Memd[a+3]
	do i = 1, np {
	    b = a + 5 * i + 3
	    yp[i] = Memd[b] * Memd[a+4]
	    xp[i] = Memd[b+1] + Memd[a+5]
	    switch (tp[i]) {
	    case GAUSS:
		gp[i] = abs (Memd[b+2] * Memd[a+6])
	    case LORENTZ:
		lp[i] = abs (Memd[b+3] * Memd[a+7])
	    case VOIGT:
		gp[i] = abs (Memd[b+2] * Memd[a+6])
		lp[i] = abs (Memd[b+3] * Memd[a+7])
	    }
	}

	call sfree (sp)
end


# DOREFIT -- Refit line profiles.  This assumes the input is very close
# to the final solution and minimizes the number of calls to the
# fitting routines.  This is intended for efficient use in the
# in computing bootstrap error estimates.

procedure dorefit (fit, x, y, s, npts, dx, nsub, y1, dy,
    xp, yp, gp, lp, tp, np, chisq)

int	fit[5]		# Fit constraints
real	x[npts]		# X data
real	y[npts]		# Y data
real	s[npts]		# Sigma data
int	npts		# Number of points
real	dx		# Pixel size
int	nsub		# Number of subpixels
real	y1		# Continuum offset
real	dy		# Continuum slope
real	xp[np]		# Profile positions
real	yp[np]		# Profile intensities
real	gp[np]		# Profile Gaussian FWHM
real	lp[np]		# Profile Lorentzian FWHM
int	tp[np]		# Profile type	
int	np		# Number of profiles
real	chisq		# Chi squared

int	i
pointer	sp, a, b
errchk	dofit1

begin
	call smark (sp)
	call salloc (a, 8 + 5 * np, TY_DOUBLE)

	# Convert positions and widths relative to first component.
	Memd[a] = dx
	Memd[a+1] = nsub
	Memd[a+2] = y1
	Memd[a+3] = dy
	Memd[a+4] = yp[1]
	Memd[a+5] = xp[1]
	Memd[a+6] = gp[1]
	Memd[a+7] = lp[1]
	do i = 1, np {
	    b = a + 5 * i + 3
	    Memd[b]   = yp[i] / Memd[a+4]
	    Memd[b+1] = xp[i] - Memd[a+5]
	    switch (tp[i]) {
	    case GAUSS:
		Memd[b+2] = gp[i] / Memd[a+6]
	    case LORENTZ:
		Memd[b+3] = lp[i] / Memd[a+7]
	    case VOIGT:
		Memd[b+2] = gp[i] / Memd[a+6]
		Memd[b+3] = lp[i] / Memd[a+7]
	    }
	    Memd[b+4] = tp[i]
	}

	# Do fit.
	call dofit1 (fit, x, y, s, npts, Memd[a], np, chisq)

	y1 = Memd[a+2]
	dy = Memd[a+3]
	do i = 1, np {
	    b = a + 5 * i + 3
	    yp[i] = Memd[b] * Memd[a+4]
	    xp[i] = Memd[b+1] + Memd[a+5]
	    switch (tp[i]) {
	    case GAUSS:
		gp[i] = abs (Memd[b+2] * Memd[a+6])
	    case LORENTZ:
		lp[i] = abs (Memd[b+3] * Memd[a+7])
	    case VOIGT:
		gp[i] = abs (Memd[b+2] * Memd[a+6])
		lp[i] = abs (Memd[b+3] * Memd[a+7])
	    }
	}

	call sfree (sp)
end


# MODEL -- Compute model.

real procedure model (x, dx, nsub, xp, yp, gp, lp, tp, np)

real	x		# X value to be evaluated
real	dx		# Pixel width
int	nsub		# Number of subpixels
real	xp[np]		# Profile positions
real	yp[np]		# Profile intensities
real	gp[np]		# Profile Gaussian FWHM
real	lp[np]		# Profile Lorentzian FWHM
int	tp[np]		# Profile type	
int	np		# Number of profiles

int	i, j
real	delta, x1, y, arg1, arg2, v, v0, u
real	ry

begin
	delta = dx / nsub
	x1 = x - (dx + delta) / 2
	y = 0.
	do j = 1, nsub {
	    x1 = x1 + delta
	    do i = 1, np {
		switch (tp[i]) {
		case GAUSS:
		    arg1 = 1.66511 * abs ((x1 - xp[i]) / gp[i])
		    if (arg1 < 5.)
			y = y + yp[i] * exp (-arg1**2)
		case LORENTZ:
		    arg2 = abs ((x1 - xp[i]) / (lp[i] / 2))
		    y = y + yp[i] / (1 + arg2**2)
		case VOIGT:
		    arg1 = 1.66511 * (x1 - xp[i]) / gp[i]
		    arg2 = 0.832555 * lp[i] / gp[i]
		    call voigt (0., arg2, v0, u)
		    call voigt (arg1, arg2, v, u)
		    y = y + yp[i] * v / v0
		}
	    }
	}
	ry = y / nsub
	return (ry)
end


# DERIVS -- Compute model and derivatives for MR_SOLVE procedure.
# This could be optimized more for the Voigt profile by reversing
# the do loops since v0 need only be computed once per line.

procedure derivs (x, a, y, dyda, na)

real	x		# X value to be evaluated
double	a[na]		# Parameters
real	y		# Function value
double	dyda[na]	# Derivatives
int	na		# Number of parameters

int	i, j, nsub
real	dx, dx1, delta, x1, wg, wl, arg1, arg2, I0, dI, c, u, v, v0

begin
	dx = a[1]
	nsub = a[2]
	delta = dx / nsub
	dx1 = .1 * delta
	x1 = x - (dx + delta) / 2
	y = 0.
	do i = 1, na
	   dyda[i] = 0.
	do j = 1, nsub {
	    x1 = x1 + delta
	    y = y + a[3] + a[4] * x1
	    dyda[3] = dyda[3] + 1.
	    dyda[4] = dyda[4] + x1
	    do i = 9, na, 5 {
		switch (a[i+4]) {
		case GAUSS:
		    I0 = a[5] * a[i]
		    wg = a[7] * a[i+2]
		    arg1 = 1.66511 * (x1 - a[6] - a[i+1]) / wg
		    if (abs (arg1) < 5.) {
			dI = exp (-arg1**2)
			c = I0 * dI * arg1
			y = y + I0 * dI
			dyda[5] = dyda[5] + a[i] * dI
			dyda[6] = dyda[6] + c / wg
			dyda[7] = dyda[7] + c * arg1 / a[7]
			dyda[i] = dyda[i] + a[5] * dI
			dyda[i+1] = dyda[i+1] + c / wg
			dyda[i+2] = dyda[i+2] + c * arg1 / a[i+2]
		    }
		case LORENTZ:
		    I0 = a[5] * a[i]
		    wl = (a[8] * a[i+3] / 2)
		    arg2 = (x1 - a[6] - a[i+1]) / wl
		    dI = 1 / (1 + arg2**2)
		    c = 2 * I0 * dI * dI * arg2
		    y = y + I0 * dI
		    dyda[5] = dyda[5] + a[i] * dI
		    dyda[6] = dyda[6] + c / wl
		    dyda[8] = dyda[8] + c * arg2 / a[8]
		    dyda[i] = dyda[i] + a[5] * dI
		    dyda[i+1] = dyda[i+1] + c / wl
		    dyda[i+3] = dyda[i+3] + c * arg2 / a[i+3]
		case VOIGT:
		    a[7] = max (dx1, abs(a[7]))
		    a[8] = max (dx1, abs(a[8]))
		    a[i+2] = max (1E-6, abs(a[i+2]))
		    a[i+3] = max (1E-6, abs(a[i+3]))

		    I0 = a[5] * a[i]
		    wg = a[7] * a[i+2]
		    wl = a[8] * a[i+3]
		    arg1 = 1.66511 * (x1 - a[6] - a[i+1]) / wg
		    arg2 = 0.832555 * wl / wg
		    call voigt (0., arg2, v0, u)
		    call voigt (arg1, arg2, v, u)
		    v = v / v0; u = u / v0
		    dI = (1 - v) / (v0 * SQRTOFPI)
		    c = 2 * I0 * arg2
		    y = y + I0 * v
		    dyda[5] = dyda[5] + a[i] * v
		    dyda[6] = dyda[6] + 2 * c * (arg1 * v - arg2 * u) / wl
		    dyda[7] = dyda[7] +
			c * (dI + arg1 * (arg1 / arg2 * v - 2 * u)) / a[7]
		    dyda[8] = dyda[8] + c * (arg1 * u - dI) / a[8]
		    dyda[i] = dyda[i] + a[5] * v
		    dyda[i+1] = dyda[i+1] + 2 * c * (arg1 * v - arg2 * u) / wl
		    dyda[i+2] = dyda[i+2] +
			c * (dI + arg1 * (arg1 / arg2 * v - 2 * u)) / a[i+2]
		    dyda[i+3] = dyda[i+3] + c * (arg1 * u - dI) / a[i+3]
		}
	    }
	}
	y = y / nsub
	do i = 1, na
	   dyda[i] = dyda[i] / nsub
end


# DOFIT1 -- Perform nonlinear iterative fit for the specified parameters.
# This uses the Levenberg-Marquardt method from NUMERICAL RECIPES.

procedure dofit1 (fit, x, y, s, npts, a, nlines, chisq)

int	fit[5]		# Fit constraints
real	x[npts]		# X data
real	y[npts]		# Y data
real	s[npts]		# Sigma data
int	npts		# Number of points
real	a[ARB]		# Fitting parameters
int	nlines		# Number of lines
real	chisq		# Chi squared

int	i, np, nfit
real	mr, chi2
pointer	sp, flags, ptr
errchk	mr_solve

begin
	# Number of terms is 5 for each line plus common background, center,
	# intensity and widths.  Also the pixel size and number of subpixels.

	np = 5 * nlines + 8

	call smark (sp)
	call salloc (flags, np, TY_INT)
	ptr = flags

	# Background.
	switch (fit[BKG]) {
	case SINGLE:
	    Memi[ptr] = 3
	    Memi[ptr+1] = 4
	    ptr = ptr + 2
	}

	# Peaks.
	switch (fit[INT]) {
	case SINGLE:
	    Memi[ptr] = 5
	    ptr = ptr + 1
	case INDEP:
	    do i = 1, nlines {
		Memi[ptr] = 5 * i + 4
		ptr = ptr + 1
	    }
	}

	# Positions.
	switch (fit[POS]) {
	case SINGLE:
	    Memi[ptr] = 6
	    ptr = ptr + 1
	case INDEP:
	    do i = 1, nlines {
		Memi[ptr] = 5 * i + 5
		ptr = ptr + 1
	    }
	}

	# Gaussian FWHM.
	switch (fit[GAU]) {
	case SINGLE:
	    Memi[ptr] = 7
	    ptr = ptr + 1
	case INDEP:
	    do i = 1, nlines {
		Memi[ptr] = 5 * i + 6
		ptr = ptr + 1
	    }
	}

	# Lorentzian FWHM.
	switch (fit[LOR]) {
	case SINGLE:
	    Memi[ptr] = 8
	    ptr = ptr + 1
	case INDEP:
	    do i = 1, nlines {
		Memi[ptr] = 5 * i + 7
		ptr = ptr + 1
	    }
	}

	nfit = ptr - flags
	mr = -1.
	i = 0
	chi2 = MAX_REAL
	repeat {
	    call mr_solve (x, y, s, npts, a, Memi[flags], np, nfit, mr, chisq)
	    if (chi2 - chisq > 0.0001)
		i = 0
	    else
		i = i + 1
	    chi2 = chisq
	} until (i == 5)

	mr = 0.
	call mr_solve (x, y, s, npts, a, Memi[flags], np, nfit, mr, chisq)

	call sfree (sp)
end


# MR_SOLVE -- Levenberg-Marquardt nonlinear chi square minimization.
#
# Use the Levenberg-Marquardt method to minimize the chi squared of a set
# of paraemters.  The parameters being fit are indexed by the flag array.
# To initialize the Marquardt parameter, MR, is less than zero.  After that
# the parameter is adjusted as needed.  To finish set the parameter to zero
# to free memory.  This procedure requires a subroutine, DERIVS, which
# takes the derivatives of the function being fit with respect to the
# parameters.  There is no limitation on the number of parameters or
# data points.  For a description of the method see NUMERICAL RECIPES
# by Press, Flannery, Teukolsky, and Vetterling, p523.

procedure mr_solve (x, y, s, npts, params, flags, np, nfit, mr, chisq)

real	x[npts]			# X data array
real	y[npts]			# Y data array
real	s[npts]			# Sigma data array
int	npts			# Number of data points
real	params[np]		# Parameter array
int	flags[np]		# Flag array indexing parameters to fit
int	np			# Number of parameters
int	nfit			# Number of parameters to fit
real	mr			# MR parameter
real	chisq			# Chi square of fit

int	i
real	chisq1
pointer	new, a1, a2, delta1, delta2

errchk	mr_invert

begin
	# Allocate memory and initialize.
	if (mr < 0.) {
	    call mfree (new, TY_DOUBLE)
	    call mfree (a1, TY_DOUBLE)
	    call mfree (a2, TY_DOUBLE)
	    call mfree (delta1, TY_DOUBLE)
	    call mfree (delta2, TY_DOUBLE)

	    call malloc (new, np, TY_DOUBLE)
	    call malloc (a1, nfit*nfit, TY_DOUBLE)
	    call malloc (a2, nfit*nfit, TY_DOUBLE)
	    call malloc (delta1, nfit, TY_DOUBLE)
	    call malloc (delta2, nfit, TY_DOUBLE)

	    call amovd (params, Memd[new], np)
	    call mr_eval (x, y, s, npts, Memd[new], flags, np, Memd[a2],
	        Memd[delta2], nfit, chisq)
	    mr = 0.001
	}

	# Restore last good fit and apply the Marquardt parameter.
	call amovd (Memd[a2], Memd[a1], nfit * nfit)
	call amovd (Memd[delta2], Memd[delta1], nfit)
	do i = 1, nfit
	    Memd[a1+(i-1)*(nfit+1)] = Memd[a2+(i-1)*(nfit+1)] * (1. + mr)

	# Matrix solution.
	call mr_invert (Memd[a1], Memd[delta1], nfit)

	# Compute the new values and curvature matrix.
	do i = 1, nfit
	    Memd[new+flags[i]-1] = params[flags[i]] + Memd[delta1+i-1]
	call mr_eval (x, y, s, npts, Memd[new], flags, np, Memd[a1],
	    Memd[delta1], nfit, chisq1)

	# Check if chisq has improved.
	if (chisq1 < chisq) {
	    mr = max (EPSILONR, 0.1 * mr)
	    chisq = chisq1
	    call amovd (Memd[a1], Memd[a2], nfit * nfit)
	    call amovd (Memd[delta1], Memd[delta2], nfit)
	    call amovd (Memd[new], params, np)
	} else
	    mr = 10. * mr

	if (mr == 0.) {
	    call mfree (new, TY_DOUBLE)
	    call mfree (a1,  TY_DOUBLE)
	    call mfree (a2,  TY_DOUBLE)
	    call mfree (delta1, TY_DOUBLE)
	    call mfree (delta2, TY_DOUBLE)
	}
end


# MR_EVAL -- Evaluate curvature matrix.  This calls procedure DERIVS.

procedure mr_eval (x, y, s, npts, params, flags, np, a, delta, nfit, chisq)

real	x[npts]			# X data array
real	y[npts]			# Y data array
real	s[npts]			# Sigma data array
int	npts			# Number of data points
double	params[np]		# Parameter array
int	flags[np]		# Flag array indexing parameters to fit
int	np			# Number of parameters
double	a[nfit,nfit]		# Curvature matrix
double	delta[nfit]		# Delta array
int	nfit			# Number of parameters to fit
double	chisq			# Chi square of fit

int	i, j, k
double	ymod, dy, dydpj, dydpk, sig2i
pointer	sp, dydp

begin
	call smark (sp)
	call salloc (dydp, np, TY_DOUBLE)

	do j = 1, nfit {
	   do k = 1, j
	       a[j,k] = 0.
	    delta[j] = 0.
	}

	chisq = 0.
	do i = 1, npts {
	    call derivs (x[i], params, ymod, Memd[dydp], np)
	    if (IS_INDEF(ymod))
		next
	    sig2i = 1. / (s[i] * s[i])
	    dy = y[i] - ymod
	    do j = 1, nfit {
		dydpj = Memd[dydp+flags[j]-1] * sig2i
		delta[j] = delta[j] + dy * dydpj
		do k = 1, j {
		    dydpk = Memd[dydp+flags[k]-1]
		    a[j,k] = a[j,k] + dydpj * dydpk
		}
	    }
	    chisq = chisq + dy * dy * sig2i
	}

	do j = 2, nfit
	    do k = 1, j-1
		a[k,j] = a[j,k]

	call sfree (sp)
end
	    

# MR_INVERT -- Solve a set of linear equations using Householder transforms.

procedure mr_invert (a, b, n)

double	a[n,n]		# Input matrix and returned inverse
double	b[n]		# Input RHS vector and returned solution
int	n		# Dimension of input matrices

int	krank
real	rnorm
pointer	sp, h, g, ip

begin
	call smark (sp)
	call salloc (h, n, TY_DOUBLE)
	call salloc (g, n, TY_DOUBLE)
	call salloc (ip, n, TY_INT)

	call hfti (a, n, n, n, b, n, 1, 1E-10, krank, rnorm,
	    Memd[h], Memd[g], Memi[ip])

	call sfree (sp)
end

# Feb 27 1997	New file
